#include <iostream>
#include <fstream>
#include <opencv2/highgui/highgui.hpp>
#include <algorithm>
#include <limits>

#include "commands/create_per_segment_estimates.h"
#include "veksler/vekslersegmentation.h"
#include "bayesian/bayesianestimator.h"
#include "gamutmapping/gamutmapping2destimator.h"
#include "gamutmapping/gamutmapping3destimator.h"
#include "grayworld/grayworldestimator.h"
#include "baseline/nothingestimator.h"
#include "baseline/meanestimator.h"
#include "evaluation/error2.h"
#include "evaluation/evaluationreport.h"
#include "common/statistics.h"
#include "img_read.h"
#include "db_descriptor.h"

#include "fusion/boosting/gradientboostingfusion.h"
#include "fusion/randomtrees/randomtreesfusion.h"

namespace illumestimators {

CommandCreatePerSegmentEstimates::CommandCreatePerSegmentEstimates()
 : vole::Command(
		"cpse",
		config,
		"Christian Riess",
		"christian.riess@informatik.uni-erlangen.de")
{
}

int CommandCreatePerSegmentEstimates::execute() {
	if (config.saveFile.length() > 0) {
		if (config.inputFile.length() < 1) { std::cerr << "input list (-I) must be set; aborted." << std::endl; return 1; }

		if (!config.onlyGW) {
			if ((config.gamut2d_1s_file.length() < 1) || (config.bayesian_el_file.length() < 1) || (config.gamut3d_1s_file.length() < 1)) {
				std::cerr << "please pass -G, -F and -B; at least one was missing, aborted." << std::endl;
				return 1;
			}
		} else {
			if ((config.gamut2d_1s_file.length() > 0) || (config.bayesian_el_file.length() > 0) || (config.gamut3d_1s_file.length() > 0)) {
				std::cerr << "WARNING: Gamut2D, Gamut3D and Bayesian Color Constancy are not used (-g set)." << std::endl;
			}
		}

		std::vector<IlluminantEstimator*> estimators;

		estimators.push_back(new NothingEstimator());

		if (!config.onlyGW) {
			estimators.push_back(getGamut2D(config.gamut2d_1s_file));
			estimators.push_back(getBayesian(config.bayesian_el_file));
			estimators.push_back(getGamut3D(config.gamut3d_1s_file));
		}

		estimators.push_back(getGw(0, -1, 0));
		estimators.push_back(getGw(0, 1, 0));
		estimators.push_back(getGw(0, 2, 0));
		estimators.push_back(getGw(0, -1, 1));
		estimators.push_back(getGw(0, 1, 1));
		estimators.push_back(getGw(0, 2, 1));
		estimators.push_back(getGw(1, -1, 1));
		estimators.push_back(getGw(1, 1, 1));
		estimators.push_back(getGw(1, 2, 1));
		estimators.push_back(getGw(2, -1, 1));
		estimators.push_back(getGw(2, 1, 1));
		estimators.push_back(getGw(2, 2, 1));

		EvaluationReport estimatesReport;
		EvaluationReport errorsReport;

		std::map<std::string, std::vector<double> > errors;

		rbase::DbDescriptor dsc(config.inputFile.c_str());
		if (!dsc.isValid) { std::cerr << "unable to open list file " << config.inputFile << std::endl; return 1; }

		int res_count = 0;
		for (int i = 0; i < dsc.nScenes; ++i) {
			std::cout << "seg: " << dsc.segmentation[i];
			// recover segmentation
			cv::Mat_<unsigned char> origMask = cv::imread(dsc.mask[i], 0);
			cv::Mat_<cv::Vec3b> segImage = cv::imread(dsc.segmentation[i]);
			std::vector<superpixels::Superpixel> superpixels;
			superpixels::SuperpixelSegmentation::imageToSuperpixels(segImage, superpixels);

			for (int j = 0; j < dsc.picturesPerScene[i]; ++j) {

				std::vector<cv::Mat_<cv::Vec3d> > preprocessedImages(estimators.size());
				std::vector<cv::Mat_<unsigned char> > preprocessedMasks(estimators.size());
				cv::Mat_<cv::Vec3d> gt = IlluminantEstimator::cleanlyReadImage(dsc.gtFiles[i][j], "rgb");
				std::cout << std::endl << "reading gt: " << dsc.gtFiles[i][j] << " , input file: " << dsc.normFiles[i][j] << std::endl;

				// note: breaks when no 16 bit images are used
				cv::Mat_<cv::Vec3d> raw_norm = iread::IRead::img_read_dl(dsc.normFiles[i][j], "rgb", dsc.dark_level[i], dsc.saturation_level[i]);

				for (unsigned int e = 0; e < estimators.size(); ++e) {
					estimators[e]->preprocessImage(raw_norm, origMask, preprocessedImages[e], preprocessedMasks[e]);
				}

				std::cout <<  dsc.gtFiles[i][j] << "  - -  " <<  dsc.normFiles[i][j] << std::endl;

				// for each segment:
				for (unsigned int s = 0; s < superpixels.size(); ++s) {
					bool bad_ground_truth = false;
					if (s % 10 == 0) std::cout << "section " << i << " image " << j << " superpixel " << s << std::endl;
					if (superpixels[s].coordinates.size() < 10) continue;
					cv::Vec3d gt_col = gt[superpixels[s].coordinates[0].y][superpixels[s].coordinates[0].x];
					if (gt_col[0] + gt_col[1] + gt_col[2] < 0.1) continue;

					if ((int)results.size() <= res_count) {
						results.push_back(std::vector<double>());
					} else {
						results[res_count] = std::vector<double>();
					}
					results[res_count].push_back(i); // first entry: scene index

					// store ground truth in result matrix
					for (int c = 0; c < 3; ++c) {
						if ((gt_col[c] != gt_col[c]) || std::isinf(gt_col[c])) bad_ground_truth = true;
						results[res_count].push_back(gt_col[c]);
					}

					if (bad_ground_truth) continue;

					// for each estimator
					// perform estimation, store in result matrix
					for (unsigned int e = 0; e < estimators.size(); ++e) {
						Illum estimate = estimators[e]->estimateIlluminant(preprocessedImages[e], superpixels[s], preprocessedMasks[e]);
						bool rescue_me = false;
						for (int c = 0; c < 3; ++c) {
							if ((estimate[c] != estimate[c]) || std::isinf(estimate[c])) rescue_me = true;
						}

						if (rescue_me) {
							for (int c = 0; c < 3; ++c) results[res_count].push_back(1.0 / 3);
						} else {
							for (int c = 0; c < 3; ++c) results[res_count].push_back(estimate[c]);
						}

						{
							IlluminantEstimator* estimator = estimators[e];

							std::stringstream ss;
							ss <<"section " << i << " image " << j << " superpixel " << s;

							Illum groundTruth(gt_col[2], gt_col[1], gt_col[0]);

							if (rescue_me) estimate = Illum(1.0 / 3, 1.0 / 3, 1.0 / 3);

							double error = estimate.angularErrorDeg(groundTruth);

							errors[estimator->name()].push_back(error);

							estimatesReport.set(ss.str(), "Ground truth", groundTruth);
							estimatesReport.setRowHeader(ss.str(), ss.str());
							estimatesReport.setColumnHeader("Ground truth", "Ground truth");

							estimatesReport.set(ss.str(), estimators[e]->name(), estimate);
							estimatesReport.setRowHeader(ss.str(), ss.str());
							estimatesReport.setColumnHeader(estimator->name(), estimator->name());

							errorsReport.set(ss.str(), estimators[e]->name(), error);
							errorsReport.setRowHeader(ss.str(), ss.str());
							errorsReport.setColumnHeader(estimator->name(), estimator->name());
						}
					}

					estimatesReport.save(config.output_path + "/estimates_single.csv");
					errorsReport.save(config.output_path + "/errors_single.csv");

					if (!bad_ground_truth) { res_count++; } else { std::cout << "nba" << std::endl; }
				}
			}

			std::cout << std::endl;
			std::cout << "number of results: " << res_count << std::endl;
			std::cout << std::endl;
		}

		EvaluationReport resultReport;

		for (std::map<std::string, std::vector<double> >::iterator it = errors.begin(); it != errors.end(); it++) {
			const double mean = Statistics::mean((*it).second);
			const double median = Statistics::median((*it).second);
			const double rms = Statistics::rms((*it).second);
			const double max = Statistics::max((*it).second);

			resultReport.setColumnHeader("mean", "Mean");
			resultReport.setColumnHeader("median", "Median");
			resultReport.setColumnHeader("rms", "RMS");
			resultReport.setColumnHeader("max", "Max");
			resultReport.setRowHeader((*it).first, (*it).first);

			resultReport.set((*it).first, "mean", mean);
			resultReport.set((*it).first, "median", median);
			resultReport.set((*it).first, "rms", rms);
			resultReport.set((*it).first, "max", max);
		}

		resultReport.save(config.output_path + "/results_single.csv");

		if (config.saveFile.length() > 0) {
			std::cout << "for comparison, the dimensions: " << results.size() << " x " << results[0].size() << std::endl;
			std::cout << "for comparison, the first 20 values: " << std::endl;
			for (unsigned int i = 0; i < std::min<unsigned int>(results[0].size(), 20); ++i) {
				std::cout << results[0][i] << "\t";
			}
			std::cout << std::endl;
			save(config.saveFile);
		}
	} else {
		if ((config.output_path.length() < 1) || (config.loadFile.length() < 1)) {
			std::cerr << "if not -s <saveFile> is passed, -O and -l must be set!" << std::endl;
			return 1;
		}
		if (!load(config.loadFile)) {
			std::cerr << "Unable to load the intermediate result file " << config.loadFile << ", aborted." << std::endl;
			return 1;
		}

		std::cout << "for comparison, the dimensions: " << results.size() << " x " << results[0].size() << std::endl;
		std::cout << "for comparison, the first 20 values: " << std::endl;
		for (unsigned int i = 0; i < std::max<unsigned int>(results[0].size(), 20); ++i) {
			std::cout << results[0][i] << "\t";
		}
		std::cout << std::endl;

		
		std::vector<std::vector<double> > single_est_errors((results[0].size()-4)/3);
		for (unsigned int i = 0; i < results.size(); ++i) {
			cv::Vec3d gt(results[i][1], results[i][2], results[i][3]);
			for (unsigned int j = 0; j < single_est_errors.size(); ++j) {
				int idx = 4+j*3;
				cv::Vec3d est(results[i][idx], results[i][idx+1], results[i][idx+2]); 
				single_est_errors[j].push_back(Error2::toAngularErrorDeg(gt, est));
			}
		}
		std::cout << std::endl;
		std::cout << "single best results: " << std::endl;
		for (unsigned int i = 0; i < single_est_errors.size(); ++i) {
			sort(single_est_errors[i].begin(), single_est_errors[i].end());
			std::cout << "entry " << i << "\tbest: " << single_est_errors[i][0] << ", worst: "
				<< single_est_errors[i][single_est_errors[i].size()-1] << " and median: "
				<< single_est_errors[i][single_est_errors[i].size()/2] << std::endl;
		}
		std::cout << std::endl;

		EvaluationReport estimatesReport;
		EvaluationReport errorsReport;

		std::map<std::string, std::vector<double> > errors;

		std::vector<double> error_gbf;
		std::vector<double> error_rtf;

		int nScenes = countScenes();

		for (int set = 0; set < nScenes; ++set) {
			std::cout << "set: " << set << std::endl;

			// create matrices for regression
			std::vector<std::vector<cv::Vec3d> > trainingEstimates;
			std::vector<cv::Vec3d> trainingIlluminants;

			std::vector<std::vector<cv::Vec3d> > testingEstimates;
			std::vector<cv::Vec3d> gtIlluminants;

			get_training_and_testing_for_set(set, trainingEstimates, trainingIlluminants, testingEstimates, gtIlluminants);

			std::cout << "training gradient boosting" << std::endl;
			GradientBoostingFusion gbf;
			gbf.train(trainingEstimates, trainingIlluminants);

			std::cout << "testing gradient boosting" << std::endl;
			for (unsigned int i = 0; i < testingEstimates.size(); ++i) {
				Illum result = gbf.estimateIlluminant(testingEstimates[i]);
				Illum gt(gtIlluminants[i][2], gtIlluminants[i][1], gtIlluminants[i][0]); // assumes bgr order;

				std::stringstream ss;
				ss << "set: " << set << " estimate " << i;

				double error = result.angularErrorDeg(gt);

				errors[gbf.name()].push_back(error);

				estimatesReport.set(ss.str(), "Ground truth", gt);
				estimatesReport.setRowHeader(ss.str(), ss.str());
				estimatesReport.setColumnHeader("Ground truth", "Ground truth");

				estimatesReport.set(ss.str(), gbf.name(), result);
				estimatesReport.setRowHeader(ss.str(), ss.str());
				estimatesReport.setColumnHeader(gbf.name(), gbf.name());

				errorsReport.set(ss.str(), gbf.name(), error);
				errorsReport.setRowHeader(ss.str(), ss.str());
				errorsReport.setColumnHeader(gbf.name(), gbf.name());
			}

			std::cout << "training random forests" << std::endl;
			RandomTreesFusion rtf;
			rtf.train(trainingEstimates, trainingIlluminants);

			std::cout << "testing random forests" << std::endl;
			for (unsigned int i = 0; i < testingEstimates.size(); ++i) {
				Illum result = rtf.estimateIlluminant(testingEstimates[i]);
				Illum gt(gtIlluminants[i][2], gtIlluminants[i][1], gtIlluminants[i][0]); // assumes bgr order;

				std::stringstream ss;
				ss << "set: " << set << " estimate " << i;

				double error = result.angularErrorDeg(gt);

				errors[rtf.name()].push_back(error);

				estimatesReport.set(ss.str(), "Ground truth", gt);
				estimatesReport.setRowHeader(ss.str(), ss.str());
				estimatesReport.setColumnHeader("Ground truth", "Ground truth");

				estimatesReport.set(ss.str(), rtf.name(), result);
				estimatesReport.setRowHeader(ss.str(), ss.str());
				estimatesReport.setColumnHeader(rtf.name(), rtf.name());

				errorsReport.set(ss.str(), rtf.name(), error);
				errorsReport.setRowHeader(ss.str(), ss.str());
				errorsReport.setColumnHeader(rtf.name(), rtf.name());
			}

			std::cout << "done" << std::endl << std::endl;
		}

		estimatesReport.save(config.output_path + "/estimates_fusion.csv");
		errorsReport.save(config.output_path + "/errors_fusion.csv");

		EvaluationReport resultReport;

		for (std::map<std::string, std::vector<double> >::iterator it = errors.begin(); it != errors.end(); it++) {
			const double mean = Statistics::mean((*it).second);
			const double median = Statistics::median((*it).second);
			const double rms = Statistics::rms((*it).second);
			const double max = Statistics::max((*it).second);

			resultReport.setColumnHeader("mean", "Mean");
			resultReport.setColumnHeader("median", "Median");
			resultReport.setColumnHeader("rms", "RMS");
			resultReport.setColumnHeader("max", "Max");
			resultReport.setRowHeader((*it).first, (*it).first);

			resultReport.set((*it).first, "mean", mean);
			resultReport.set((*it).first, "median", median);
			resultReport.set((*it).first, "rms", rms);
			resultReport.set((*it).first, "max", max);
		}

		resultReport.save(config.output_path + "/results_fusion.csv");
	}

	return 0;
}

int CommandCreatePerSegmentEstimates::countScenes()
{
	std::set<int> scene_id;
	for (unsigned int i = 0; i < results.size(); ++i) {
		scene_id.insert((int)(results[i][0] + 0.1));
	}
	return scene_id.size();
}

void CommandCreatePerSegmentEstimates::get_training_and_testing_for_set(int set,
	std::vector<std::vector<cv::Vec3d> > &trainingEstimates, std::vector<cv::Vec3d> &trainingIlluminants,
	std::vector<std::vector<cv::Vec3d> > &testingEstimates, std::vector<cv::Vec3d> &gtIlluminants)
{
	int testEntries = (results[0].size()-4)/3;
	int testingCount = 0;
	int trainingCount = 0;
	for (unsigned int i = 0; i < results.size(); ++i) {
		if (fabs((results[i][0] - set)) < 0.1) { // testing
			gtIlluminants.push_back(cv::Vec3d(results[i][1], results[i][2], results[i][3]));
			testingEstimates.push_back(cv::vector<cv::Vec3d>());
			for (int bla = 0; bla < testEntries; ++bla) {
				testingEstimates[testingCount].push_back(cv::Vec3d(results[i][bla*3+4], results[i][bla*3+5], results[i][bla*3+6]));
			}
			testingCount++;
		} else {
			trainingIlluminants.push_back(cv::Vec3d(results[i][1], results[i][2], results[i][3]));
			trainingEstimates.push_back(cv::vector<cv::Vec3d>());
			for (int bla = 0; bla < testEntries; ++bla) {
				trainingEstimates[trainingCount].push_back(cv::Vec3d(results[i][bla*3+4], results[i][bla*3+5], results[i][bla*3+6]));
			}
			trainingCount++;
		}
	}
}

int CommandCreatePerSegmentEstimates::keyToNum(std::string &key) {
	if (key.compare("chalk") == 0) return 0;
	if (key.compare("figures") == 0) return 1;
	if (key.compare("fruits") == 0) return 2;
	if (key.compare("rabbit") == 0) return 3;
	return -1;
}


IlluminantEstimator *CommandCreatePerSegmentEstimates::getGamut2D(std::string db_file) {
	GamutMapping2DEstimator *est = new GamutMapping2DEstimator();
	est->load(db_file);
	return est;
}

IlluminantEstimator *CommandCreatePerSegmentEstimates::getGamut3D(std::string db_file) {
	GamutMapping3DEstimator *est = new GamutMapping3DEstimator();
	est->load(db_file);
	return est;
}

IlluminantEstimator *CommandCreatePerSegmentEstimates::getBayesian(std::string db_file) {
	BayesianEstimator *est = new BayesianEstimator();
	est->load(db_file);
	return est;
}

IlluminantEstimator *CommandCreatePerSegmentEstimates::getGw(int n, int p, int s) {
	return new GrayWorldEstimator(n, p, s);
}

void CommandCreatePerSegmentEstimates::compute_illuminant_error(std::vector<cv::Vec3d> &illuminants, std::vector<cv::Vec3d> &gt_values, cv::Mat_<double> &errors)
{
	int nMetrics = 5;
	errors = cv::Mat_<double>(nMetrics, illuminants.size());
	for (int i = 0; i < (int)illuminants.size(); ++i) {
		errors[0][i] = Error2::toAngularErrorDeg(illuminants[i], gt_values[i]);
		errors[1][i] = Error2::toRgError(illuminants[i], gt_values[i]);
		errors[2][i] = Error2::toRMSE(illuminants[i], gt_values[i]);
		errors[3][i] = Error2::toRError(illuminants[i], gt_values[i]);
		errors[4][i] = Error2::toGError(illuminants[i], gt_values[i]);
	}
}

bool CommandCreatePerSegmentEstimates::load(std::string filename) {
	cv::FileStorage	fs(filename, cv::FileStorage::READ);
	if (!fs.isOpened()) {
		std::cerr << "unable to open file storage " << filename << std::endl;
		return false;
	}

	std::string name = fs["name"];

	if (name != "EstimateCollection") {
		std::cerr << "wrong method name: expected EstimateCollection" << name << std::endl;
		return false;
	}

	cv::Mat_<double> readMe;
	fs["singleResults"] >> readMe;

	// unpack readme into results;
	results = std::vector<std::vector<double> >(readMe.rows);
	for (int y = 0; y < readMe.rows; ++y) {
		results[y] = std::vector<double>(readMe.cols);
		for (int x = 0; x < readMe.cols; ++x) {
			results[y][x] = readMe[y][x];
		}
	}

	return true;


}

bool CommandCreatePerSegmentEstimates::save(std::string filename) {
	cv::FileStorage	fs(filename, cv::FileStorage::WRITE);

	if (!fs.isOpened()) {
		std::cerr << "unable to open file storage " << filename << std::endl;
		return false;
	}

	fs << "name" << "EstimateCollection";

	// repackage everything in a matrix
	int ysize = results.size();
	if (ysize == 0) return false;
	int xsize = results[0].size();
	
	cv::Mat_<double> writeMe(ysize, xsize);
	int skip = 0;
	for (int y = 0; y < ysize; ++y) {
		if ((int) results[y].size() != writeMe.cols) { skip++; continue; }

		for (int x = 0; x < xsize; ++x) {
			writeMe[y-skip][x] = results[y][x];
		}
	}
	writeMe = writeMe.rowRange(0, writeMe.rows-skip);
	fs << "singleResults" << writeMe;

	return true;
}

void CommandCreatePerSegmentEstimates::printShortHelp() const {
	std::cout << "	Create Per Segment Illuminant Estimates" << std::endl;
}

void CommandCreatePerSegmentEstimates::printHelp() const {
	std::cout << "Create Per Segment Illuminant Estimates" << std::endl;
}

}
