#include "evalcommand.h"
#include "metadata/metadatastorage.h"
#include "pipeline/single_illuminant_pipeline.h"
#include "evaluation/crossvalidation.h"
#include "gamutmapping/gamutmapping2destimator.h"
#include "gamutmapping/gamutmapping3destimator.h"
#include "bayesian/bayesianestimator.h"
#include "grayworld/grayworldestimator.h"
#include "baseline/meanestimator.h"
#include "baseline/nothingestimator.h"
#include "fusion/randomtrees/randomtreesfusion.h"
#include "fusion/boosting/gradientboostingfusion.h"
#include "evaluation/evaluationreport.h"
#include "common/statistics.h"
#ifndef BOOST_FILESYSTEM_VERSION
#define BOOST_FILESYSTEM_VERSION 3
#endif // BOOST_FILESYSTEM_VERSION
#include <boost/filesystem.hpp>

namespace illumestimators {

EvalCommand::EvalCommand():
	Command("eval",
		config,
		"Michael Bleier",
		"michael.bleier@mb.stud.uni-erlangen.de"
	)
{
}

int EvalCommand::execute()
{
	MetadataStorage storage(config.metaFiles);

	std::vector<std::string> imageFiles;

	for (unsigned int i = 0; i < config.inputFiles.size(); i++) {
		std::string filename = config.inputFiles[i];
		boost::filesystem::path path(filename);

		if (path.extension().string().compare(".txt") == 0) {
			std::ifstream in(filename.c_str(), std::ifstream::in);

			while (in.good()) {
				std::string line;
				std::getline(in, line);

				if (line.size() > 0) {
					if (boost::filesystem::path(line).is_absolute()) {
						imageFiles.push_back(line);
					} else {
						imageFiles.push_back(path.parent_path().string() + "/" + line);
					}
				}
			}
		} else {
			imageFiles.push_back(filename);
		}
	}

	std::cout << "habe " << imageFiles.size() << " Bilder" << std::endl;
	std::vector<CrossValidation::Set> sets = CrossValidation(imageFiles.size(), CrossValidation::LEAVEONEOUT).sets();

	std::vector<IlluminantEstimator*> estimators;

#if 1
	// Baseline
	estimators.push_back(new NothingEstimator());
	estimators.push_back(new MeanEstimator());
#endif

#if 0
	// Standard set Gray World
	estimators.push_back(new GrayWorldEstimator(0, -1, 0));
	estimators.push_back(new GrayWorldEstimator(0, 1, 0));
	estimators.push_back(new GrayWorldEstimator(1, 1, 1));
	estimators.push_back(new GrayWorldEstimator(2, 1, 1));
#endif

#if 1
	// Extended set Gray World
	for (int p = -1; p <= 10; p++) {
		if (p == 0) continue;
		for (double sigma = 0; sigma <= 2; sigma = sigma + 1) {
			estimators.push_back(new GrayWorldEstimator(0, p, sigma));
			if (sigma == 0) continue;
			estimators.push_back(new GrayWorldEstimator(1, p, sigma));
			estimators.push_back(new GrayWorldEstimator(2, p, sigma));
		}
	}
#endif

#if 1
	// Gamut mapping
	estimators.push_back(new GamutMapping3DEstimator(0, 0, config.verbosity, true));
	estimators.push_back(new GamutMapping2DEstimator(GamutMapping2DEstimator::MEAN, 0, 0, config.verbosity, true));
	estimators.push_back(new GamutMapping2DEstimator(GamutMapping2DEstimator::MEDIAN, 0, 0, config.verbosity, true));
#endif

#if 0
	// Bayesian
	estimators.push_back(new BayesianEstimator(BayesianEstimator::MIN_AVERAGE_LOSS_EUCLIDEAN, config.verbosity, true));
#endif

	EvaluationReport estimatesReport;
	EvaluationReport errorsReport;

	std::map<std::string, std::vector<double> > errors;

	for (unsigned int j = 0; j < sets.size(); j++) {
		CrossValidation::Set set = sets[j];

		std::vector<std::string> trainFiles;
		for (unsigned int k = 0; k < set.training().size(); k++) {
			trainFiles.push_back(imageFiles[set.training()[k]]);
		}

		std::vector<std::string> testFiles;
		for (unsigned int k = 0; k < set.testing().size(); k++) {
			testFiles.push_back(imageFiles[set.testing()[k]]);
		}

		for (unsigned int i = 0; i < estimators.size(); i++) {
			IlluminantEstimator* estimator = estimators[i];

			if (!SingleIlluminantPipeline::runTraining(*estimator, trainFiles, storage, config.verbosity)) {
				std::cerr << "unable to train " << estimator->name() << ", aborted." << std::endl;
				return 1;
			}

			for (unsigned int k = 0; k < testFiles.size(); k++) {
				Illum estimate;
				double error;

				if (!SingleIlluminantPipeline::runEstimator(estimate, error, *estimator, testFiles[k], storage, config.verbosity)) {
					std::cerr << "EvalCommand::execute(): SingleIlluminantPipeline::runEstimator failed, aborted." << std::endl;
					return -1;
				}

				errors[estimator->name()].push_back(error);

				estimatesReport.set(testFiles[k], estimator->name(), estimate);
				estimatesReport.setRowHeader(testFiles[k], testFiles[k]);
				estimatesReport.setColumnHeader(estimator->name(), estimator->name());

				errorsReport.set(testFiles[k], estimator->name(), error);
				errorsReport.setRowHeader(testFiles[k], testFiles[k]);
				errorsReport.setColumnHeader(estimator->name(), estimator->name());
			}
		}

		estimatesReport.save(config.outputDirectory + "/estimates.txt");
		errorsReport.save(config.outputDirectory + "/errors.txt");
	}

	EvaluationReport resultReport;

	for (std::map<std::string, std::vector<double> >::iterator it = errors.begin(); it != errors.end(); it++) {
		const double mean = Statistics::mean((*it).second);
		const double median = Statistics::median((*it).second);
		const double rms = Statistics::rms((*it).second);
		const double max = Statistics::max((*it).second);

		resultReport.setColumnHeader("mean", "Mean");
		resultReport.setColumnHeader("median", "Median");
		resultReport.setColumnHeader("rms", "RMS");
		resultReport.setColumnHeader("max", "Max");
		resultReport.setRowHeader((*it).first, (*it).first);

		resultReport.set((*it).first, "mean", mean);
		resultReport.set((*it).first, "median", median);
		resultReport.set((*it).first, "rms", rms);
		resultReport.set((*it).first, "max", max);
	}

	resultReport.save(config.outputDirectory + "/result.txt");

	return 0;
}

void EvalCommand::printShortHelp() const
{
	std::cout << "	Illuminant estimator evaluation." << std::endl;
}

void EvalCommand::printHelp() const
{
	std::cout << "Illuminant estimator evaluation." << std::endl;
	std::cout << "\"Input file\" are a list of file names, each containing again a white space-delimited list of file names." << std::endl;

}

} // namespace illumestimators
